# An implementation of the Stackless Python "stackless" module using greenlets.

import greenlet

class Scheduler:
    def __init__(self):
        self.taskletsByID = {}
        self.tasklets = []
        self.currentIdx = 0
        self.currentTasklet = None

        # Note what greenlet we are started in, this is the main thread of execution.
        self.greenlet = greenlet.getcurrent()        

    def Run(self, timeout, remove=False):
        # Ensure that tasklets are only scheduled by the main thread of execution.
        if greenlet.getcurrent() is not self.greenlet:
            raise RuntimeError("Run must be run from the main tasklet.")

        if len(self.tasklets):
            # The current tasklet index can change through whatever logic
            # happens as a result of the run() call.  So we want to hold
            # a reference to it in order to ensure we can provide it if needs be.
            try:
                # Run the current tasklet until it yields or exits.
                self.currentTasklet = self.tasklets[self.currentIdx]
                self.currentTasklet.run()

                # We only want to remove the current tasklet or move onto the next
                # one if the current one is still in the current position.
                if len(self.tasklets) and self.tasklets[self.currentIdx] is self.currentTasklet:
                    if remove:
                        # Remove the tasklet.  This should correct the index for the next in line.
                        self.RemoveTasklet(self.currentTasklet)
                    else:
                        # The tasklet was run with an explicit request to leave it to be scheduled
                        # again.  Just make the next tasklet the current one.
                        self.currentIdx += 1
                        if self.currentIdx >= len(self.tasklets):
                            self.currentIdx = 0
            finally:
                self.currentTasklet = None

        return None

    def GetCurrentTasklet(self):
        return self.currentTasklet

    def GetIndexedTasklet(self):
        return self.tasklets[self.currentIdx]

    def DecrementIndex(self):
        self.currentIdx -= 1
        if self.currentIdx < 0:
            self.currentIdx = len(self.tasklets) - 1

    # Make the current tasklet last and add a new one first.
    def InsertTasklet(self, t):
        if not self.AddTasklet(t, last=False):
            # There are probably better ways..
            self.RemoveTasklet(t)
            self.AddTasklet(t, last=False)

    # Add a new tasklet after all the others.
    def AddTasklet(self, t, last=True):
        if t.blocked:
            raise RuntimeError("You cannot run a blocked tasklet")

        if self.taskletsByID.has_key(t.id):
            return False

        self.taskletsByID[t.id] = t
        if last:
            if len(self.tasklets) != 0 and self.currentIdx != 0:
                self.tasklets.insert(self.currentIdx, t)
                self.currentIdx += 1            
            else:
                self.tasklets.append(t)
                if len(self.tasklets) == 1:
                    self.currentIdx = 0
        else:
            if len(self.tasklets) and self.currentIdx < len(self.tasklets) - 1:
                self.tasklets.insert(self.currentIdx + 1, t)
                self.currentIdx += 1
            else:
                self.tasklets.append(t)
                self.currentIdx = len(self.tasklets) - 1
        self.UpdateRunCount()
        return True

    def RemoveTasklet(self, t, idx=None):
        if not self.taskletsByID.has_key(t.id):
            return

        del self.taskletsByID[t.id]
        if idx is None:
            idx = self.tasklets.index(t)
        del self.tasklets[idx]
        if self.currentIdx > idx:
            self.currentIdx -= 1
        if self.currentIdx >= len(self.tasklets):
            self.currentIdx = 0
        self.UpdateRunCount()

    def UpdateRunCount(self):
        global runcount
        runcount = len(self.tasklets)

    def Yield(self):
        """Switch back to this scheduler.  This ensures that the scheduler
        finishes its current operation and it is up to how the application
        is using the scheduler what happens next."""
        return self.greenlet.switch()

scheduler = Scheduler()
runcount = 0

def run(timeout):
    """Run the current tasklet until it yields or exits.  If it takes longer than the timeout
    period, then it will be returned.  Otherwise None will be returned.  The timeout parameter
    is ignored at this time."""
    return scheduler.Run(timeout, remove=False)

def getcurrent():
    """If a tasklet is currently being executed by the scheduler, return it."""
    return scheduler.GetCurrentTasklet()

def schedule():
    """schedule(retval=stackless.current) -- switch to the next runnable tasklet.
    The return value for this call is retval, with the current tasklet as default."""
    return scheduler.Run(0, remove=False)

def schedule_remove():
    """schedule_remove(retval=stackless.current) -- same as schedule but remove self."""
    return scheduler.Run(0, remove=True)

class channel:
    def __init__(self):
        self.balance = 0
        self.queue = []

    def send(self, value):
        """channel.send(value) -- send a value over the channel.
        If no other tasklet is already receiving on the channel, the sender will
        be blocked. Otherwise, the receiver will be activated immediately, and
        the sender is put at the end of the runnables list.
        Note that an exception instance sent will be raised at the receiver
        (see also channel.send_exception)"""
        if self.balance < 0:
            self.balance += 1
            t = self.queue.pop(0)
            t.args = (value,)
            t.blocked = 0
            scheduler.InsertTasklet(t)
        else:
            t = getcurrent()
            # Remove the tasklet from the list of running tasklets.
            t.remove()
            # Let it wait for a receiver to come along.
            self.balance += 1
            self.queue.append((t, value))
            t.blocked = 1
        return scheduler.Yield()

    def receive(self):
        """channel.receive() -- receive a value over the channel.
        If no other tasklet is already sending on the channel,
        the receiver will be blocked. Otherwise, the receiver will
        continue immediately, and the sender is put at the end of
        the runnables list."""
        if self.balance > 0:
            self.balance -= 1
            t, value = self.queue.pop(0)
            t.blocked = 0
            scheduler.AddTasklet(t)
            return value

        t = getcurrent()
        t.remove()
        self.balance -= 1
        self.queue.append(t)
        t.blocked = -1
        return scheduler.Yield()

class tasklet:
    nextID = 1

    def __init__(self, func):
        self.func = func
        self.args = ()
        self.kwargs = {}
        self.greenlet = None

        self.name = ""
        self.alive = False
        self.blocked = 0
        self.paused = False

        self.id = tasklet.nextID
        tasklet.nextID += 1

    def __repr__(self):
        return "<Tasklet id=%d name=\"%s\"" % (self.id, self.name)

    def __call__(self, *args, **kwargs):
        # Store the arguments for when we are scheduled.
        self.args = args
        self.kwargs = kwargs

        self.insert()

        return self

    def insert(self):
        """ Insert this tasklet at the end of the scheduler list, given that it isn't blocked.
        Blocked tasklets need to be reactivated by channels."""
        scheduler.AddTasklet(self)

    def remove(self):
        """Removing a tasklet from the runnables queue.
        Note: If this tasklet has a non-trivial C stack attached,
        it will be destructed when the containing thread state is destroyed.
        Since this will happen in some unpredictable order, it may cause unwanted
        side-effects. Therefore it is recommended to either run tasklets to the
        end or to explicitly kill() them.
        """
        scheduler.RemoveTasklet(self)

    def run(self):
        """ Run this tasklet, given that it isn't blocked.
            Blocked tasks need to be reactivated by channels.""" 
        if self.blocked:
            raise RuntimeError("You cannot run a blocked tasklet")

        if self.greenlet is None:
            self.alive = True
            func, args, kwargs = self.func, self.args, self.kwargs
            self.func, self.args, self.kwargs = None, (), {}
            self.greenlet = greenlet.greenlet(func)
            self.greenlet.switch(*args)
        else:
            args, kwargs = self.args, self.kwargs
            self.args, self.kwargs = (), {}
            self.greenlet.switch(*args)

        # If the greenlet is not non-zero, then it died on us.
        if not self.greenlet:
            self.alive = False
            scheduler.RemoveTasklet(self)

    def become(self, value):
        """t.become(retval) -- catch the current running frame in a tasklet.
        It is also inserted at the end of the runnables chain.
        If it is a toplevel frame (and therefore has no caller), an exception is raised.
        The function result is the tasklet itself. retval is passed to the calling frame.
        If retval is not given, the tasklet is used as default."""
        pass
        # Cannot see how to do this offhand.

    def capture(self, value):
        """t.capture(retval) -- capture the current running frame in a tasklet,
        like t.become(). In addition the tasklet is run immediately, and the
        parent tasklet is removed from the runnables and returned as the value."""
        pass
        # Cannot see how to do this offhand.
